require 'color'
require 'json'
require 'uri'

# Wrap an Altmetric data hash and provide translation to badge URIs
class AltmetricBadger
  BADGE_TYPES = %i(donut pie helix bar)
  BASE_HOST = 'altmetric-badges.a.ssl.fastly.net'

  class InvalidDataError < StandardError; end

  def initialize(score: nil, data:)
    fail InvalidDataError unless data && data.respond_to?(:sort_by)
    @data = data
    @score = score
  end

  BADGE_TYPES.each do |type|
    define_method "to_#{type}_uri" do |size: 640, svg: false, show_score: true, ssl: true|
      (ssl ? URI::HTTPS : URI::HTTP).build(
        host: BASE_HOST,
        path: '/',
        query: "style=#{ type }&score=#{ score if show_score }&types=#{ type_string }&size=#{ size }#{ '&svg' if svg }"
      )
    end
  end

  private

  attr_reader :data, :score

  SEGMENT_COUNT = 8
  DEFAULT_TYPE_STRING = '?' * SEGMENT_COUNT

  class << self
    def sources
      @sources ||= JSON.parse(
        File.read(
          File.join(__dir__, '..', 'data', 'sources.json')
        )
      )
    end

    def sources_key_map
      @sources_key_map ||= sources.map { |e| [e['key'], e] }.to_h
    end

    def sources_code_map
      @sources_code_map ||= sources.map { |e| [e['code'], e] }.to_h
    end

    def key_for_code(code)
      sources_code_map.fetch(code, {})['key']
    end

    def code_for_key(type)
      sources_key_map.fetch(type, sources_key_map[nil])['code']
    end

    def color_for_code(code)
      sources_code_map.fetch(code, sources_key_map[nil])['color']
    end
  end

  # Return string of type codes that Altmetric badge URI uses
  def type_string
    return DEFAULT_TYPE_STRING unless total_posts > 0

    @type_string ||= segments_to_string(
      apportion_segments(
        post_percentages.keys.last(SEGMENT_COUNT)
      )
    )
  end

  # Return a hash of each post type to the percentage of posts of that type
  def post_percentages
    @post_percentages ||= data.sort_by { |_, v| v }.map do |type, count|
      [self.class.code_for_key(type), (count / total_posts.to_f)]
    end.to_h
  end

  # Divide segments among all post types, reflecting their proportion, and
  # ensuring that the most popular SEGMENT_COUNT post types receive at least
  # one segment each
  def apportion_segments(segments)
    (SEGMENT_COUNT - segments.length).times do
      segments << post_percentages.max_by { |_, v| v }.first
      post_percentages[segments.last] -= (1 / SEGMENT_COUNT.to_f)
    end
    segments
  end

  # Return the total number of posts in the data hash
  def total_posts
    @total_posts ||= data.values.reduce(:+) || 0
  end

  # Sort list of segments by apparent brightness and return combined string
  def segments_to_string(segments)
    segments.sort do |segment1, segment2|
      color_difference_value(segment1, segment2)
    end.join
  end

  # Calculate an absolute difference value between two colors
  # This is pretty stupid and broken, but we need to keep the same behaviour
  # as the Altmetric API so that the donuts don't look different.
  def color_difference_value(segment1, segment2)
    color1, color2 = [segment1, segment2].map do |color|
      Color::RGB.from_html(self.class.color_for_code(color))
    end.map(&:to_hsl)

    %i(hue saturation brightness).map do |n|
      color1.send(n) - color2.send(n)
    end.reduce(:+)
  end
end
